/*
 * USB Orinoco driver
 *
 * Copyright (c) 2003 Manuel Estrada Sainz <ranty@debian.org>
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License
 * at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License version 2 (the "GPL"), in
 * which case the provisions of the GPL are applicable instead of the
 * above.  If you wish to allow the use of your version of this file
 * only under the terms of the GPL and not to allow others to use your
 * version of this file under the MPL, indicate your decision by
 * deleting the provisions above and replace them with the notice and
 * other provisions required by the GPL.  If you do not delete the
 * provisions above, a recipient may use your version of this file
 * under either the MPL or the GPL.  
 *
 * Queueing code based on linux-wlan-ng 0.2.1-pre5
 *
 * Copyright (C) 1999 AbsoluteValue Systems, Inc.  All Rights Reserved.
 *
 * 	The license is the same as above.
 *
 * Initialy based on USB Skeleton driver - 0.7
 *
 * Copyright (c) 2001 Greg Kroah-Hartman (greg@kroah.com)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation; either version 2 of
 *	the License, or (at your option) any later version.
 *
 * NOTE: The original USB Skeleton driver is GPL, but all that code is
 * gone so MPL/GPL applies.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/smp_lock.h>
#include <linux/usb.h>
#include <linux/timer.h>
#include <asm/io.h>

#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/etherdevice.h>
#include <linux/wireless.h>

#include "hermes.h"
#include "hermes_rid.h"
#include "orinoco.h"

#include "linux/firmware.h"
#include "orinoco_usb.h"
#include "orinoco_usb_fw.h"
#include "kcompat.h"

#ifdef CONFIG_USB_DEBUG
	static int debug = 1;
#else
	static int debug;
#endif

/* Use our own dbg macro */
#undef dbg
#define dbg(format, arg...) do { if (debug) printk(KERN_DEBUG __FILE__":%s: " format "\n", __FUNCTION__ , ## arg); } while (0)

#define ndbg(format, arg...)
#define dump(a, b) bridge_dump_data(__FUNCTION__, a, b)
#define ndump(a, b)

#define UNUSED __attribute__((unused))

/* Version Information */
#define DRIVER_VERSION "v0.1.4"

/* Module paramaters */
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug enabled or not");

/* Some times, the card gets stuck and stops paying
 * attention to the world until we do something to it. Setting
 * watchdog_secs to a nonzero value will set a timer on the USB
 * callbacks that will trigger within watchdog_secs seconds of
 * inactivity. The method to get the card back to work is just to send an
 * INQ_TALLIES command which should be harmless */
static int watchdog_secs = 0;
MODULE_PARM(watchdog_secs, "i");
MODULE_PARM_DESC(watchdog_secs, "0 to disable");

static int force_unsupported = 0;
MODULE_PARM(force_unsupported, "i");
MODULE_PARM_DESC(force_unsupported, "Try to handle untested hardware.");

static int default_card_variant = 0;
MODULE_PARM(default_card_variant, "i");
MODULE_PARM_DESC(default_card_variant, "Card variant to use as a first guess");
/* Most of this Device/Vendor pairs are untested */

#define USB_COMPAQ_VENDOR_ID		0x049f /* Compaq Computer Corp. */
#define USB_COMPAQ_WL215_ID		0x001f /* Compaq WL215 USB Adapter */
#define USB_COMPAQ_W200_ID		0x0076 /* Compaq W200 USB Adapter */

#define USB_MELCO_VENDOR_ID		0x0411
#define USB_BUFFALO_AIRSTA_ID		0x000B /* Buffalo airstation usb */

#define USB_LUCENT_VENDOR_ID		0x047E /* Lucent Technologies */
#define USB_LUCENT_ORINOCO_ID		0x0300 /* Lucent/Agere Orinoco USB
						  Client */
#define USB_AVAYA8_VENDOR_ID		0x0D98
#define USB_AVAYAE_VENDOR_ID		0x0D98
#define USB_AVAYA_WIRELESS_ID		0x0300 /* Avaya Wireless USB Card */

#define USB_AGERE_VENDOR_ID		0x0D4E /* Agere Systems */
#define USB_AGERE_ORINOCO_ID		0x1000

#define USB_ELSA_VERNDOR_ID		0x05CC
#define USB_ELSA_AIRLANCER_ID		0x3100 /* ELSA AirLancer USB-11 */

#define USB_LEGEND_VENDOR_ID		0x0E7C
#define USB_LEGEND_JOYNET_ID		0x0300 /* LEGEND Joynet WLAN USB Card */

#define USB_SAMSUNG_VENDOR_ID		0x04E8
#define USB_SAMSUNG_SEW2001U_ID		0x5002 /* Samsung SEW-2001u Card */
#define USB_SANSUNG_SEW2001U2_ID	0x5B11 /* Samsung SEW-2001u Card */

#define USB_IGATE_VENDOR_ID		0x0681
#define USB_IGATE_IGATE_11M_ID		0x0012 /* I-GATE 11M USB Card */

#define USB_FUJITSU_VENDOR_ID		0x0BF
#define USB_FUJITSU_E1100_ID		0x1002 /* Fujitsu Siemens Computers
						  connect2AIR WLAN E-1100 USB */
#define USB_2WIRE_VENDOR_ID		0x1630
#define USB_2WIRE_WIRELESS_ID		0xff81 /* 2Wire Wireless USB adapter */


/* table of devices that may work with this driver */
/* We will refuse to touch any untested device unless force_unsupported is set
 */
static struct usb_device_id bridge_table [] = {
	{ USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_WL215_ID) },
	{ USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_W200_ID) },
	{ USB_DEVICE(USB_MELCO_VENDOR_ID, USB_BUFFALO_AIRSTA_ID) },
	{ USB_DEVICE(USB_LUCENT_VENDOR_ID, USB_LUCENT_ORINOCO_ID) },
	{ USB_DEVICE(USB_AVAYA8_VENDOR_ID, USB_AVAYA_WIRELESS_ID) },
	{ USB_DEVICE(USB_AVAYAE_VENDOR_ID, USB_AVAYA_WIRELESS_ID) },
	{ USB_DEVICE(USB_AGERE_VENDOR_ID, USB_AGERE_ORINOCO_ID) },
	{ USB_DEVICE(USB_ELSA_VERNDOR_ID, USB_ELSA_AIRLANCER_ID) },
	{ USB_DEVICE(USB_LEGEND_VENDOR_ID, USB_LEGEND_JOYNET_ID) },
	{ USB_DEVICE_VER(USB_SAMSUNG_VENDOR_ID, USB_SAMSUNG_SEW2001U_ID,
			 0, 0) },
	{ USB_DEVICE(USB_SAMSUNG_VENDOR_ID, USB_SANSUNG_SEW2001U2_ID) },
	{ USB_DEVICE(USB_IGATE_VENDOR_ID, USB_IGATE_IGATE_11M_ID) },
	{ USB_DEVICE(USB_FUJITSU_VENDOR_ID, USB_FUJITSU_E1100_ID) },
	{ USB_DEVICE(USB_2WIRE_VENDOR_ID, USB_2WIRE_WIRELESS_ID) },
	{ USB_DEVICE(0x0D4E, 0x047A) },
	{ }					/* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, bridge_table);

/* Devices should be added to 'bridge_table' and one of
 * bridge_varian1_table or bridge_varian2_table. Devices on the
 * 'bridge_table' but not on any of 'bridge_varian1_table' or
 * 'bridge_varian2_table' will require force_unsupported=1.  */

/* Variant 1 has a PCMCIA card inside */
static struct usb_device_id bridge_varian1_table [] = {
	{ USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_WL215_ID) },
	{ USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_W200_ID) },
	{ USB_DEVICE(USB_LUCENT_VENDOR_ID, USB_LUCENT_ORINOCO_ID) },
	{ USB_DEVICE(USB_MELCO_VENDOR_ID, USB_BUFFALO_AIRSTA_ID) },
	{ }
};
/* Variant 2 is all build in a single board */
static struct usb_device_id bridge_varian2_table [] = {
	{ }
};

struct bridge_read {
	struct urb *urb;
	u8 pipe;
	int avail_status;
};
struct bridge_write {
	u8 pipe;
};
enum bridge_priv_flags {
	ORINOCO_USB_NEEDS_AC_REQ=1,
};	
struct request_queue {
	spinlock_t		lock;
	struct list_head	pending;
	struct list_head	active;
};
/* Structure to hold all of our device specific stuff */
struct bridge_priv {
	struct usb_device *udev;
	struct usb_interface *interface;
	struct net_device *net_dev;
	struct bridge_read read;
	struct bridge_write write;
	u8 reply_count;
	struct timer_list watchdog_timer;
	struct semaphore sem;
	u8 card_variant;
	u16 hermes_reg_fake [0x40];
	int (*orinoco_stop)(struct net_device *dev);
	int (*orinoco_open)(struct net_device *dev);
	u32 flags;
	
	struct request_queue ctxq;
	struct {
		u8 *buf;
		int fill;
	} bap;
};

struct request_context
{
	struct list_head	list;
	atomic_t		refcount;
	struct completion	done;		/* Signals that CTLX is dead */
	struct urb		*outurb;	/* OUT for req pkt */
	/* the RID that we will wait for */
	u16 out_rid;
	u16 in_rid;
	struct bridge_priv	*dev;
	union {
		struct bridge_req_packet out;
		struct bridge_ans_packet in;
	} *buf;
	int buf_actual_length;
	struct timer_list	reqtimer;	/* For IN(response)wait */
	struct timer_list	resptimer;	/* For OUT(request) wait */
	volatile u32		state;		/* Tracks running state */
};

enum {
	BRIDGE_CTX_START,
	BRIDGE_CTX_QUEUED,
	BRIDGE_CTX_REQ_SUBMITTED,
	BRIDGE_CTX_REQ_COMPLETE,
	BRIDGE_CTX_RESP_RECEIVED,
	BRIDGE_CTX_REQ_TIMEOUT,
	BRIDGE_CTX_REQ_FAILED,
	BRIDGE_CTX_RESP_TIMEOUT,
	BRIDGE_CTX_REQSUBMIT_FAIL,
	BRIDGE_CTX_COMPLETE,
};

static void bridge_or_reg(struct hermes *hw, u16 off, u16 bits);

static __inline__ u8 bridge_reply_inc(u8 count);
static void bridge_ctx_complete(struct request_context *ctx);

static void 
bridge_request_resptimerfn(u_long data)
{
	struct request_context	*ctx = (void *)data;
	struct bridge_priv	*dev = ctx->dev;
	unsigned long		flags;
	dbg("ENTER");
	spin_lock_irqsave(&dev->ctxq.lock, flags);
	if ( !dev->udev ) {
		spin_unlock_irqrestore(&dev->ctxq.lock, flags);
		return;
	}

	list_del_init(&ctx->list);
	ctx->state = BRIDGE_CTX_RESP_TIMEOUT;
	spin_unlock_irqrestore(&dev->ctxq.lock, flags);

	bridge_ctx_complete(ctx);
};

static int
bridge_request_context_alloc(struct request_context **ctx_p,
			     struct bridge_priv *dev, u16 out_rid, u16 in_rid)
{
	struct request_context *ctx;

	ctx = kmalloc(sizeof(*ctx), GFP_ATOMIC);
	if (!ctx) {
		return -ENOMEM;
	}
	memset(ctx, 0, sizeof(*ctx));

	ctx->buf = kmalloc(BULK_BUF_SIZE, GFP_ATOMIC);
	if (!ctx->buf) {
		kfree(ctx);
		return -ENOMEM;
	}
	ctx->outurb = usb_alloc_urb(0, GFP_ATOMIC);
	if (!ctx->outurb) {
		kfree(ctx->buf);
		kfree(ctx);
		return -ENOMEM;
	}

	*ctx_p = ctx;

	ctx->dev = dev;
	ctx->state = 0;
	ctx->out_rid = out_rid;
	ctx->in_rid = in_rid;

	atomic_set(&ctx->refcount, 1);
	init_completion(&ctx->done);
	//usb_init_urb(&ctx->outurb);
	//ctx->outurb->timeout = DEF_TIMEOUT;

	init_timer(&ctx->resptimer);
	ctx->resptimer.function = bridge_request_resptimerfn;
	ctx->resptimer.data = (unsigned long)ctx;

	init_timer(&ctx->reqtimer);
	ctx->reqtimer.function = bridge_request_resptimerfn;
	ctx->reqtimer.data = (unsigned long)ctx;
	return 0;
}
static void 
bridge_request_context_free(struct request_context *ctx)
{
	if (!ctx)
		return;
	if (atomic_dec_and_test(&ctx->refcount)) {
		ndbg("freeing %p:%p", ctx->buf, ctx);
		WARN_ON(!ctx->done.done);
		
		del_timer(&ctx->reqtimer);
		del_timer(&ctx->resptimer);
		usb_free_urb(ctx->outurb);
		kfree(ctx->buf);
		kfree(ctx);
	}
}
static void bridge_req_queue_run(struct request_queue *ctxq);

/**
 * bridge_ctx_complete:
 * Description:
 *	Note: Timers have been stopped, and ctx should be in a
 *	terminal state.
 **/
static void
bridge_ctx_complete(struct request_context *ctx)
{
	struct bridge_priv *dev = ctx->dev;
	struct net_device *net_dev = dev->net_dev;
	struct orinoco_private *priv = net_dev->priv;
	hermes_t *hw = &priv->hw;
	unsigned long		flags;

	spin_lock_irqsave(&dev->ctxq.lock, flags);
	if (dev->udev) {
		list_del(&ctx->list);

		spin_unlock_irqrestore(&dev->ctxq.lock, flags);
		bridge_req_queue_run(&dev->ctxq);
		spin_lock_irqsave(&dev->ctxq.lock, flags);
	}

	switch(ctx->state) {
	case BRIDGE_CTX_COMPLETE:
	case BRIDGE_CTX_REQSUBMIT_FAIL:
	case BRIDGE_CTX_REQ_FAILED:
	case BRIDGE_CTX_REQ_TIMEOUT:
	case BRIDGE_CTX_RESP_TIMEOUT:
		spin_unlock_irqrestore(&dev->ctxq.lock, flags);

		switch (ctx->out_rid) {
		case BRIDGE_RID_TX:
			ndbg("BRIDGE_RID_TX status:%d",
			    ctx->outurb->status);
			if (ctx->state != BRIDGE_CTX_COMPLETE ){
				bridge_or_reg(hw, HERMES_EVSTAT,
					      HERMES_EV_TXEXC);
			} else {
				bridge_or_reg(hw, HERMES_EVSTAT,
					      HERMES_EV_TX);
			}
			orinoco_interrupt(net_dev->irq, net_dev, NULL);
			netif_wake_queue(net_dev);
			break;
		}
		complete(&ctx->done);
		bridge_request_context_free(ctx);
		break;
		
	default:
		spin_unlock_irqrestore(&dev->ctxq.lock, flags);
		
		err("Called, CTLX not in terminating state.");
		/* Things are really bad if this happens. Just leak
		 * the CTLX because it may still be linked to the 
		 * queue or the OUT urb may still be active.
		 * Just leaking at least prevents an Oops or Panic.
		 */
		break;
	}
}

/**
 * bridge_req_queue_run:
 * Description:
 *	Note: Only one active CTX at any one time, because there's no
 *	other (reliable) way to match the response URB to the correct
 *	CTX.
 **/

static void
bridge_req_queue_run(struct request_queue *ctxq)
{
	unsigned long		flags;

	spin_lock_irqsave(&ctxq->lock, flags);

	if (list_empty(&ctxq->active) && !list_empty(&ctxq->pending)) {
		struct request_context	*head;
		int			result;

		head = list_entry(ctxq->pending.next, struct request_context,
				  list);

		if (!head->dev->udev)
			goto unlock;

		/* We need to split this off to avoid a race condition */
		list_move_tail(&head->list, &ctxq->active);

		if (head->state == BRIDGE_CTX_QUEUED) {
			result = usb_submit_urb(head->outurb, GFP_ATOMIC);
			if (result) {
				head->state = BRIDGE_CTX_REQSUBMIT_FAIL;

				spin_unlock_irqrestore(&ctxq->lock, flags);

				err("Fatal, failed to submit command urb."
				    " error=%d\n", result);

				bridge_ctx_complete(head);
				goto done;
			}

			head->state = BRIDGE_CTX_REQ_SUBMITTED;
			
			head->resptimer.expires = jiffies + DEF_TIMEOUT*2;
			add_timer(&head->resptimer);
			
			head->reqtimer.expires = jiffies + DEF_TIMEOUT;
			add_timer(&head->reqtimer);
		}
	}
	
unlock:
	spin_unlock_irqrestore(&ctxq->lock, flags);
	
done:
	return;
}

static void 
bridge_req_enqueue_run(struct request_queue *ctxq, struct request_context *ctx)
{
	unsigned long		flags;

	spin_lock_irqsave(&ctxq->lock, flags);

	if (!ctx->dev->udev) {
		spin_unlock_irqrestore(&ctxq->lock, flags);
		goto done;
	}
	atomic_inc(&ctx->refcount);
	list_add_tail(&ctx->list, &ctxq->pending);
	ctx->state = BRIDGE_CTX_QUEUED;

	spin_unlock_irqrestore(&ctxq->lock, flags);

	bridge_req_queue_run(ctxq);

	done:
	return;
}

static void
bridge_request_out_callback(struct urb *urb, struct pt_regs *pt_regs)
{
	unsigned long flags;
	unsigned int state;
	struct request_context *ctx = urb->context;
	struct bridge_priv *dev = ctx->dev;

	del_timer(&ctx->reqtimer);
	spin_lock_irqsave(&dev->ctxq.lock, flags);
	state = ctx->state;

	if(watchdog_secs)
		mod_timer(&dev->watchdog_timer, jiffies+watchdog_secs*HZ);

	if ( urb->status == 0 ) {
		switch ( state ) {
		case BRIDGE_CTX_REQ_SUBMITTED:
			if (ctx->in_rid) {
				ctx->state = BRIDGE_CTX_REQ_COMPLETE;
				spin_unlock_irqrestore(&dev->ctxq.lock, flags);
				/* reply URB still pending */
				break;
			}
			/* fall through */
		case BRIDGE_CTX_RESP_RECEIVED:
			/* IN already received before this OUT-ACK */
			ctx->state = BRIDGE_CTX_COMPLETE;
			spin_unlock_irqrestore(&dev->ctxq.lock, flags);
			bridge_ctx_complete(ctx);
			break;

		default:
			spin_unlock_irqrestore(&dev->ctxq.lock, flags);
			err("Illegal CTLX success state(0x%x, %d)"
			    " in OUT URB\n", state, urb->status);
			break;
		}	
	} else {
		/* If someone cancels the OUT URB then its status
		 * should be either -ECONNRESET or -ENOENT.
		 */
		switch ( state ) {
		case BRIDGE_CTX_REQ_SUBMITTED:
			del_timer(&ctx->resptimer);
			/* fall through */

		case BRIDGE_CTX_RESP_RECEIVED:
			ctx->state = BRIDGE_CTX_REQ_FAILED;
			/* fall through */

		case BRIDGE_CTX_REQ_FAILED:
		case BRIDGE_CTX_REQ_TIMEOUT:
			spin_unlock_irqrestore(&dev->ctxq.lock, flags);

			bridge_ctx_complete(ctx);
			break;

		default:
			spin_unlock_irqrestore(&dev->ctxq.lock, flags);

			err("Illegal CTLX error state(0x%x, %d) in OUT URB",
			    state, urb->status);
			break;
		}
	}
}
static int
bridge_request_in_callback(struct bridge_priv *dev, struct urb *urb)
{
	struct bridge_ans_packet *ans = urb->transfer_buffer;
	struct request_context *ctx = NULL;
	unsigned int state = BRIDGE_CTX_START; /* for clobbering */
	unsigned long flags;

	/* Find the CTLX on the active queue that requested this URB */
	spin_lock_irqsave(&dev->ctxq.lock, flags);
	if ( dev->udev ) {
		struct list_head *item;

		list_for_each(item, &dev->ctxq.active) {
			struct request_context *c;
			int reply_count;

			c = list_entry(item, struct request_context, list);
			reply_count = bridge_reply_inc(c->buf->out.reply_count);
			if ((ans->reply_count == reply_count)
			    && (ans->hermes_rid == c->in_rid)) {
				ctx = c;
				state = ctx->state;
				break;
			}
			dbg("Skiped (0x%x/0x%x) (%d/%d)", ans->hermes_rid,
			    c->in_rid, ans->reply_count, reply_count);
		}
	}

	if ( ctx == NULL ) {
		spin_unlock_irqrestore(&dev->ctxq.lock, flags);
		warn("Could not match IN URB(0x%x,%d) to CTLX - ignored",
		     ans->hermes_rid, urb->status);
		bridge_req_queue_run(&dev->ctxq);
		return -ENOENT;
	} else {
		/* The data we want is in the in buffer, exchange */
		urb->transfer_buffer = ctx->buf;
		ctx->buf = (void *)ans;
		ctx->buf_actual_length = urb->actual_length;
	}

	switch ( state ) {
	case BRIDGE_CTX_REQ_SUBMITTED:
		/* We have received our response URB before
		 * our request has been acknowledged. Do NOT
		 * destroy our CTX yet, because our OUT URB
		 * is still alive ...
		 */
		ctx->state = BRIDGE_CTX_RESP_RECEIVED;
		spin_unlock_irqrestore(&dev->ctxq.lock, flags);

		/* Stop the intimer */
		del_timer(&ctx->resptimer);

		/* Copy the buffer to ctx */
		//memcpy(&ctx->buf->in, ans, sizeof(ctx->buf->in));

		/* Let the machine continue running. */
		break;

	case BRIDGE_CTX_REQ_COMPLETE:
		/* This is the usual path: our request
		 * has already been acknowledged, and
		 * we have now received the reply.
		 */
		ctx->state = BRIDGE_CTX_COMPLETE;
		spin_unlock_irqrestore(&dev->ctxq.lock, flags);

		/* Stop the intimer */
		del_timer(&ctx->resptimer);

		/* Copy the buffer to ctx */
		//memcpy(&ctx->buf->in, ans, sizeof(ctx->buf->in));

		/* Call the completion handler */
		bridge_ctx_complete(ctx);
		break;

	default:
		spin_unlock_irqrestore(&dev->ctxq.lock, flags);

		/* Throw this CTX away and try submitting another */
		del_timer(&ctx->reqtimer);
		del_timer(&ctx->resptimer);
		ctx->outurb->transfer_flags |= URB_ASYNC_UNLINK;
		usb_unlink_urb(ctx->outurb);
		bridge_req_queue_run(&dev->ctxq);

		warn("Matched IN URB, CTX in invalid state(0x%x). Discarded.",
		     state);
		break;
	} /* switch */
	return 0;
}



static void 
bridge_req_ctx_wait(struct bridge_priv *dev, struct request_context *ctx)
{
	unsigned long	flags;

	spin_lock_irqsave(&dev->ctxq.lock, flags);
	switch(ctx->state) { 
	case BRIDGE_CTX_COMPLETE:
	case BRIDGE_CTX_REQSUBMIT_FAIL:
	case BRIDGE_CTX_REQ_TIMEOUT:
	case BRIDGE_CTX_REQ_FAILED:
	case BRIDGE_CTX_RESP_TIMEOUT:
		dbg("Already done, skipping sleep.");
		/* fall through */
	case BRIDGE_CTX_START:
		spin_unlock_irqrestore(&dev->ctxq.lock, flags);
		break;
	default:
		spin_unlock_irqrestore(&dev->ctxq.lock, flags);
		dbg("Sleeping...");
		if (in_interrupt()) {
			while(!ctx->done.done)
				udelay(1000);
		} else {
			wait_event_interruptible(ctx->done.wait,
						 ctx->done.done);
		}
		break;
	}
}

/* local function prototypes */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void * bridge_probe(struct usb_device *dev, unsigned int ifnum,
			   const struct usb_device_id *id);
static void bridge_disconnect(struct usb_device *dev, void *ptr);
#else
static int bridge_probe(struct usb_interface *interface,
			const struct usb_device_id *id);
static void bridge_disconnect(struct usb_interface *interface);
#endif


static void bridge_bulk_in_callback (struct urb *urb, struct pt_regs *);

static inline void
bridge_dump_data(const char *function, const void *data, int size);

static char *rid_name(u16 rid);

/* lock to protect the minor_table structure */
static DECLARE_MUTEX (minor_table_mutex);
/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver orinoco_driver = {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,19))
	.owner		= THIS_MODULE,
#endif
	.name		= "Orinoco USB",
	.probe		= bridge_probe,
	.disconnect	= bridge_disconnect,
	.id_table	= bridge_table,
};

/*
 * Orinoco helper functions
 */

static __inline__ u16 build_crc(u8 *data)
{
	u16 crc = 0;
	int i;
	for (i=0; i<8; i++){
		crc = (crc<<1) + data[i];
	}
	return crc;
}

static __inline__ u8 bridge_reply_inc(u8 count)
{
	if (count < 0x7F)
		return count+1;
	else
		return 1;
}

/**
 * fill_bridge_req:
 * 
 * if data == NULL and length > 0 the data is assumed to be already in
 * the target buffer and only the header is filled.
 *
 */
static int 
fill_bridge_req(struct bridge_req_packet *req, u16 length, u16 rid,
		const void *data, u16 x04, u8 reply_count)
{
	int total_size = sizeof(*req)+length;

	BUG_ON(total_size > BULK_BUF_SIZE);

	req->x00 = 0x0210;
	req->reply_count = reply_count;
	req->x03 = 0x00;
	req->x04 = x04;
	req->x05 = 0x00;
	req->size= length + 4;
	req->crc = build_crc((u8*)req);
	req->hermes_len = (length+3)/2;
	req->hermes_rid = rid;
	if(data)
		memcpy(req->data, data, length);
	return total_size;
}

static int bridge_submit_in_urb(struct bridge_priv *dev,
				struct request_context *ctx)
{
	int retval = 0;
	void *cur_buf = dev->read.urb->transfer_buffer;
	ndbg("ENTER");
	if(dev->read.urb->status == -EINPROGRESS){
		dbg("urb busy, not resubmiting");
		retval = -EBUSY;
		goto exit;
	}
	if(ctx){
		void *tmp = ctx->buf;
		ctx->buf = cur_buf;
		cur_buf = tmp;
		ctx->buf_actual_length = dev->read.urb->actual_length;
	}
	dev->read.avail_status = dev->read.urb->status;
	usb_fill_bulk_urb(dev->read.urb, dev->udev,
			  usb_rcvbulkpipe(dev->udev, dev->read.pipe),
			  cur_buf, BULK_BUF_SIZE,
			  bridge_bulk_in_callback, dev);
	dev->read.urb->timeout = 0; //300*HZ;
	dev->read.urb->transfer_flags = 0; //USB_ASYNC_UNLINK;
        retval = usb_submit_urb(dev->read.urb, GFP_ATOMIC);
        if (retval) {
		err("%s submit failed %d", __FUNCTION__, retval);
		goto exit;
        }
exit:
	ndbg("EXIT");
	return retval;
}
static int bridge_remove_in_urb(struct bridge_priv *dev)
{
	int retval = 0;
	ndbg("ENTER");
	if (dev->read.urb->status != -EINPROGRESS){
		dbg("no urb to remove");
		retval = -ENOENT;
		goto exit;
	}
	retval = usb_unlink_urb(dev->read.urb);
exit:	
	ndbg("EXIT");
	return retval;
}
static __inline__ int ezusb_8051_cpucs(struct bridge_priv *dev, u8 reset)
{
	if(!dev->udev){
		err("%s: !dev->udev", __FUNCTION__);
		return -EFAULT;
	}
	return usb_control_msg(dev->udev,
			       usb_sndctrlpipe(dev->udev, 0),
			       EZUSB_REQUEST_FW_TRANS,
			       USB_TYPE_VENDOR|USB_RECIP_DEVICE|USB_DIR_OUT,
			       EZUSB_CPUCS_REG, 0x0,
			       &reset, sizeof(reset),
			       DEF_TIMEOUT);
}

static int bridge_reset(struct bridge_priv *dev)
{
	int retval = 0;
	ndbg("ENTER");
	retval = ezusb_8051_cpucs(dev, 1);
	if (retval < 0){
		err("Failed to reset");
		goto exit;
	}

	retval = ezusb_8051_cpucs(dev, 0);
	if (retval < 0){
		err("Failed to unreset");
		goto exit;
	}
exit:
	ndbg("EXIT");
	return retval;
}

static int find_fw_variant_offset(struct ez_usb_fw *fw)
{
	int offset;
	offset = *(short *)(fw->code+1);
	offset = be16_to_cpu(offset);
	for(; offset < fw->size; offset++){
		if (memcmp("\x02\x03\x58", &fw->code[offset], 3)==0)
			break;
	}
	if (offset > (fw->size-4)){
		dbg("couldn't find card_variant_offset");
		return -1;
	}
	fw->card_variant_offset = offset + 3;
	dbg("card_variant_offset = 0x%04X", fw->card_variant_offset);
	return 0;
}

static int bridge_firmware_download(struct bridge_priv *dev,
				    struct ez_usb_fw *fw)
{
	int retval, addr;
	int packet_size = dev->udev->descriptor.bMaxPacketSize0;
	int variant_offset = fw->card_variant_offset%packet_size;
	int variant_base = fw->card_variant_offset - variant_offset;
        u8 *transfer_buffer =  kmalloc (packet_size, GFP_KERNEL);

	dbg("ENTER");
	ndbg("packet size: %d", packet_size);
        if (!transfer_buffer) {
                err("orinoco_usb: kmalloc(%d) failed.", packet_size);
                retval = -ENOMEM;
		goto fail;
        }

	retval = ezusb_8051_cpucs(dev, 1);
	if(retval < 0)
		goto fail;
	for(addr = 0; addr < fw->size; addr += packet_size){
		if((addr >= 0x100) && (addr < 0x300)){
			/* 0x100-0x300 should be left alone, it
			 * contains card specific data, like USB
			 * enumeration information */
			ndbg("Skiping 0x100-0x300 (0x%04X)", fw_rec->Address);
			continue;
		}
		memcpy (transfer_buffer, &fw->code[addr], packet_size);
		if(addr == variant_base){
			dbg("Patching card_variant byte 0x%02X at 0x%04X",
			    dev->card_variant, fw->card_variant_offset);
			transfer_buffer[variant_offset] = dev->card_variant;
		}
		retval = usb_control_msg(dev->udev,
					 usb_sndctrlpipe(dev->udev, 0),
					 EZUSB_REQUEST_FW_TRANS,
					 USB_TYPE_VENDOR|USB_RECIP_DEVICE
					 |USB_DIR_OUT,
					 addr, 0x0,
					 transfer_buffer, packet_size,
					 DEF_TIMEOUT);

		if(retval < 0)
			goto fail;
	}
	retval = ezusb_8051_cpucs(dev, 0);
	goto exit;
fail:
	err("orinoco_usb: fw upload: retval=%d", retval);
exit:
	if(transfer_buffer)
		kfree (transfer_buffer);
	dbg("EXIT");
	return retval;
}
static int bridge_access_ltv(struct bridge_priv *dev,
			     struct request_context *ctx,
			     u16 length, const void *data, u16 x04,
			     void *ans_buff, int ans_size,
			     u16 *ans_length, const int locked)
{
	struct net_device *net_dev = dev->net_dev;
	struct orinoco_private *priv = net_dev->priv;
	hermes_t *hw = &priv->hw;
	int req_size;
	int retval = 0;
	unsigned long flags=0;

	ndbg("ENTER");
	BUG_ON(!ctx);

	if(!dev->udev){
		dbg("Device disconnected");
		return -ENODEV;
	}

	if(dev->read.urb->status != -EINPROGRESS){
		err("%s: in urb not pending", __FUNCTION__);
	}

	BUG_ON(locked);

	spin_lock_irqsave(&hw->lock, flags);
	/* need to be atomic to get reply_count right */
	req_size = fill_bridge_req(&ctx->buf->out, length, ctx->out_rid, data,
				   x04, dev->reply_count);
	usb_fill_bulk_urb(ctx->outurb, dev->udev, 
			  usb_sndbulkpipe(dev->udev, dev->write.pipe),
			  &ctx->buf->out, req_size,
			  bridge_request_out_callback, ctx);

	if (ctx->in_rid)
		dev->reply_count = bridge_reply_inc(dev->reply_count);

	bridge_req_enqueue_run(&dev->ctxq, ctx);

	spin_unlock_irqrestore(&hw->lock, flags);

	if (ctx->in_rid) {
		int hw_unavailable = priv->hw_unavailable;
#warning hw_unavailable is probably not right, but it works for now
		if(!hw_unavailable)
			spin_unlock_bh(&priv->lock);

		bridge_req_ctx_wait(dev, ctx);

		if(!hw_unavailable)
			spin_lock_bh(&priv->lock);
	}

	switch (ctx->state){
	case BRIDGE_CTX_COMPLETE:
                retval = ctx->outurb->status;
		break;

	case BRIDGE_CTX_QUEUED:
	case BRIDGE_CTX_REQ_SUBMITTED:
		if(!ctx->in_rid)
			break;
	default:
		err("Non specific state %d", ctx->state);
		/* fall though */
	case BRIDGE_CTX_REQ_TIMEOUT:
	case BRIDGE_CTX_REQ_FAILED:
	case BRIDGE_CTX_RESP_TIMEOUT:
	case BRIDGE_CTX_REQSUBMIT_FAIL:
		err("orinoco_usb: access failed(%d)"
		    " reply_count %d (resetting)",
		    ctx->state, dev->reply_count);
		dev->reply_count = 0;
		if (ctx->state == BRIDGE_CTX_REQ_TIMEOUT
		    || ctx->state == BRIDGE_CTX_RESP_TIMEOUT) {
                        err("orinoco_usb: ctx timed out");
                        retval = -ETIMEDOUT;
                } else {
			err("orinoco_usb: ctx failed");
                        retval = -EFAULT;
                }
		goto exit;
		break;
	}
	if (ctx->in_rid) {
		struct bridge_ans_packet *ans = &ctx->buf->in;
		ndump((__u8 *)ans, ctx->buf_actual_length);
		if(((ans->hermes_len+1)*2+10) != ctx->buf_actual_length)
			err("%s: hermes_len=%d actual_length=%d", __FUNCTION__,
			    ans->hermes_len, ctx->buf_actual_length);
		if(ans_buff)
			memcpy(ans_buff, ans->data,
			       min_t(int, ctx->buf_actual_length-sizeof(*ans),
				     ans_size));
		if(ans_length)
			*ans_length = ans->hermes_len;
	}
exit:
	bridge_request_context_free(ctx);
	ndbg("EXIT");
	return retval;
}

int bridge_write_ltv(hermes_t *hw, int bap, u16 rid,
		     u16 length, const void *data) 
{
	struct bridge_priv *dev = hw->priv;
	u16 x04 = 0x02;
	struct request_context *ctx;
	int err;
	
	if (length == 0)
		return -EINVAL;
	ndbg("length: %d", length);
	length = HERMES_RECLEN_TO_BYTES(length);
	ndbg("length: %d", length);
	if (length == 0){
		/* On memory mapped devices
		 * HERMES_RID_CNFGROUPADDRESSES can be set to be
		 * empty, but the USB bridge doesn't like it */
		ndbg("zero size data, make them think that we did it");
		return 0;
	}
	err = bridge_request_context_alloc(&ctx, dev, rid, BRIDGE_RID_ACK);
	if (err)
		return err;
	
	ndbg("0x%04X", rid);
	if (rid == BRIDGE_RID_TX)
		x04 = 0x01;
	
	return bridge_access_ltv(dev, ctx, length, data, x04,
				 NULL, 0, NULL, 0);
}

static int bridge_read_ltv(hermes_t *hw, int bap, u16 rid, unsigned bufsize,
			   u16 *length, void *buf)
{
	struct bridge_priv *dev = hw->priv;
	struct request_context *ctx;
	int err;
	ndbg("0x%04X", rid);

	if ( (bufsize < 0) || (bufsize % 2) )
		return -EINVAL;

	err = bridge_request_context_alloc(&ctx, dev, rid, rid);
	if (err)
		return err;
	
	return bridge_access_ltv(dev, ctx, 0, NULL, 0x02,
				 buf, bufsize, length, 0);
}
static int bridge_docmd_wait(hermes_t *hw, u16 cmd, u16 parm0,
			     hermes_response_t * resp, int locked)
{
	struct bridge_priv *dev = hw->priv;
	struct request_context *ctx;
	int err;
	
	struct {
		u16 cmd;
		u16 parm0;
		u16 parm1;
		u16 parm2;
	} data = {
		.cmd = cmd,
		.parm0 = parm0,
		.parm1 = 0,
		.parm2 = 0,
	};
	dbg("0x%04X, parm0 0x%04X", cmd, parm0);
	ndump((void *)&data, sizeof(data));
	err = bridge_request_context_alloc(&ctx, dev,
					   BRIDGE_RID_DOCMD, BRIDGE_RID_ACK);
	if (err)
		return err;
	
	return bridge_access_ltv(dev, ctx, sizeof(data), &data, 0x02,
				 NULL, 0, NULL, locked);
}
static char *reg_name(u16 off)
{
	switch (off){
	case HERMES_INTEN:
		return "HERMES_INTEN";
	case HERMES_EVSTAT:
		return "HERMES_EVSTAT";
	case HERMES_EVACK:
		return "HERMES_EVACK";
	case HERMES_TXCOMPLFID:
		return "HERMES_TXCOMPLFID";
	case HERMES_ALLOCFID:
		return "HERMES_ALLOCFID";
	case HERMES_SWSUPPORT0:
		return "HERMES_SWSUPPORT0";
	case HERMES_INFOFID:
		return "HERMES_INFOFID";
	default:
		return "";
	}
}
static char *rid_name(u16 rid)
{
	switch (rid){
	case HERMES_INQ_LINKSTATUS:
		return "HERMES_INQ_LINKSTATUS";
	case HERMES_INQ_TALLIES:
		return "HERMES_INQ_TALLIES";
	case HERMES_INQ_SCAN:
		return "HERMES_INQ_SCAN";
	case BRIDGE_RID_RX:
		return "BRIDGE_RID_RX";
	case BRIDGE_RID_TX:
		return "BRIDGE_RID_TX";
	default:
		return "";
	}
}
static u16 bridge_read_reg(struct hermes *hw, u16 off)
{
	struct bridge_priv *dev = hw->priv;
	u16 aux;
	switch (off){
	case HERMES_RXFID:
		aux = BRIDGE_RID_RX;
		break;
	default:
		dbg("off: 0x%04X(%s)", off, reg_name(off));
	case HERMES_EVSTAT:
	case HERMES_SWSUPPORT0:
	case HERMES_ALLOCFID:
	case HERMES_INFOFID:
		aux = dev->hermes_reg_fake[off];
	}
	ndbg("off: 0x%04X(%s) val: 0x%04X", off, reg_name(off), aux);
	return aux;
}
static void bridge_write_reg(struct hermes *hw, u16 off, u16 val)
{
	struct bridge_priv *dev = hw->priv;
	u16 aux;
	ndbg("off: 0x%04X(%s) val: 0x%04X", off, reg_name(off), val);
	switch (off){
	case HERMES_EVACK:
		aux = bridge_read_reg(hw, HERMES_EVSTAT);
		aux &= ~val;
		bridge_write_reg(hw, HERMES_EVSTAT, aux);
		break;
	default:
		err("orinoco_usb: unknown offset: off: 0x%04X(%s) val: 0x%04X",
		    off, reg_name(off), val);
	case HERMES_EVSTAT:
	case HERMES_INTEN:
	case HERMES_SWSUPPORT0:
	case HERMES_ALLOCFID:
	case HERMES_TXCOMPLFID:
	case HERMES_INFOFID:
		dev->hermes_reg_fake[off] = val;
	}
	return;
}
static void bridge_or_reg(struct hermes *hw, u16 off, u16 bits)
{
	u16 aux = bridge_read_reg(hw, off);
	aux |= bits;
	bridge_write_reg(hw, off, aux);
}
static int bridge_bap_pread(struct hermes *hw, int bap,
			    void *buf, unsigned len,
			    u16 id, u16 offset)
{
	struct bridge_priv *dev = hw->priv;
	struct bridge_ans_packet *ans =
		(void *)dev->read.urb->transfer_buffer;
	int actual_length = dev->read.urb->actual_length;

	ndbg("bap: %d, id: 0x%04X(%s), offset: 0x%04X, len: %d, size: %d",
	    bap, id, rid_name(id), offset, len, actual_length);
	if (id != BRIDGE_RID_RX){
		ndbg("id == %d != BRIDGE_RID_RX", id);
	}
	if(offset == 0){
		ndump((void *)ans, actual_length);
	}
	switch(id){
	case HERMES_INQ_LINKSTATUS:
	case HERMES_INQ_TALLIES:
	case HERMES_INQ_SCAN:
		/* FIXME: this should be the default */
		if ((sizeof(*ans)+offset+len-4) > actual_length){
			err("orinoco_usb: bap read beyond end actual_length");
			return -EFAULT;
		}
		/* Should include the LENGTH/TYPE part */
		memcpy(buf, (ans->data-4)+offset, len);
		ndbg("dev->read.bap.offset: %d", dev->read.bap.offset);
		break;
	default:
		if ((sizeof(*ans)+offset+len) > actual_length){
			err("orinoco_usb: bap read beyond end actual_length");
			return -EFAULT;
		}
		memcpy(buf, ans->data+offset, len);
	}
	ndump(buf, len);
	return 0;
}

static int bridge_bap_pwrite(struct hermes *hw, int bap, const void *buf,
			     unsigned len, u16 id, u16 offset)
{
	struct bridge_priv *dev = hw->priv;
	struct bridge_req_packet *req = (void *)dev->bap.buf;

	dbg("bap: %d, id: 0x%04X, offset: %d, len: %d",
	    bap, id, offset, len);

	BUG();

	if(offset == 0){
		memset(dev->bap.buf, 0, BULK_BUF_SIZE);
		ndbg("clearing bap");
	}
	memcpy(req->data+offset, buf, len);
	dev->bap.fill = max_t(int, sizeof(*req)+offset+len, dev->bap.fill);
	return 0;
}
static int bridge_xmit(hermes_t *hw, int bap, struct memvect vect[],
			      u16 fid)
{
	int i, err;
	struct bridge_priv *dev = hw->priv;
	struct hermes_tx_descriptor *desc = vect[0].buf;
	struct request_context *ctx;
	u8 *buf;
	int eat = HERMES_802_3_OFFSET - 2;
	int tx_size = 0;
	
	err = bridge_request_context_alloc(&ctx, dev, BRIDGE_RID_TX, 0);
	if (err)
		goto exit;
	
	memset(ctx->buf, 0, BULK_BUF_SIZE);
	buf = ctx->buf->out.data;

	/* EZ-USB firmware takes care of TX_OK and TX_EX, and setting
	 * them seams to confuse it */
 	desc->tx_control &= ~(HERMES_TXCTRL_TX_OK | HERMES_TXCTRL_TX_EX);
	memcpy(buf, &desc->tx_control, sizeof(desc->tx_control));

	for (i=1; vect[i].buf; i++) {
		memcpy(buf + vect[i].offset - eat,
		       vect[i].buf, vect[i].size);
		tx_size = vect[i].offset - eat + vect[i].size;
	}
	ndbg("tx_size: %d", tx_size);
	WARN_ON(tx_size == 0);
	err = bridge_access_ltv(dev, ctx, tx_size, NULL, 0x01,
				NULL, 0, NULL, 0);
exit:
	return err;
}

static int bridge_allocate(struct hermes *hw, u16 size, u16 *fid)
{
	struct bridge_priv *dev = hw->priv;
	struct net_device *net_dev = dev->net_dev;

	ndbg("size: %d", size);
	*fid = BRIDGE_RID_TX;
	hermes_write_regn(hw, ALLOCFID, BRIDGE_RID_TX);
	bridge_or_reg(hw, HERMES_EVSTAT, HERMES_EV_ALLOC);
	orinoco_interrupt(net_dev->irq, net_dev, NULL);

	return 0;
}

static int bridge_init(hermes_t *hw)
{
	struct bridge_priv *dev = hw->priv;
	int retval = 0;
	dbg("ENTER");
	if(in_interrupt()){
		err("bridge_init called from interrupt");
		retval = -EFAULT;
		goto exit;
	}
	if(!dev){
		err("NULL dev");
		retval = -EFAULT;
		goto exit;
	}
	dev->reply_count=0;
	/* Write the MAGIC number on the simulated registers to keep
	 * orinoco.c happy */
	hermes_write_regn(hw, SWSUPPORT0, HERMES_MAGIC);
	retval = bridge_reset(dev);
	if (retval < 0)
		goto exit;
	dbg("sending control message");
	retval = usb_control_msg(dev->udev,
				 usb_sndctrlpipe(dev->udev, 0),
				 EZUSB_REQUEST_TRIGER,
				 USB_TYPE_VENDOR|USB_RECIP_DEVICE|USB_DIR_OUT,
				 0x0, 0x0,
				 NULL, 0,
				 DEF_TIMEOUT);
	if (retval < 0){
		err("EZUSB_REQUEST_TRIGER failed retval %d", retval);
		goto exit;
	}
	if (dev->flags && ORINOCO_USB_NEEDS_AC_REQ){
		dbg("Sending EZUSB_REQUEST_TRIG_AC");
		retval = usb_control_msg(dev->udev,
					 usb_sndctrlpipe(dev->udev, 0),
					 EZUSB_REQUEST_TRIG_AC,
					 USB_TYPE_VENDOR|USB_RECIP_DEVICE
					 |USB_DIR_OUT,
					 0x00FA, 0x0,
					 NULL, 0,
					 DEF_TIMEOUT);
		if (retval < 0){
			err("EZUSB_REQUEST_TRIG_AC failed retval %d", retval);
			goto exit;
		}
	}
	bridge_remove_in_urb(dev);
	bridge_submit_in_urb(dev, NULL);

	retval = bridge_write_ltv(hw, 0, BRIDGE_RID_INIT1,
				  HERMES_BYTES_TO_RECLEN(2), "\x10\x00");
	if (retval < 0){
		err("orinoco_usb: BRIDGE_RID_INIT1 retval %d", retval);
		goto exit;
	}
	retval = hermes_docmd_wait(hw, HERMES_CMD_INIT, 0, NULL);
	if (retval < 0){
		err("orinoco_usb: HERMES_CMD_INIT retval %d", retval);
		goto exit;
	}

exit:
	if(retval < 0){
		if (bridge_reset(dev) < 0){
			err("orinoco_usb: ERROR RESETTING DEVICE");
		}
	}
	dbg("EXIT");
	return retval;
}

static int bridge_netdev_stop(struct net_device *net_dev)
{
	struct orinoco_private *priv = net_dev->priv;
	struct bridge_priv *dev = priv->card;

	if(watchdog_secs)
		del_timer(&dev->watchdog_timer);
	return dev->orinoco_stop(net_dev);
}
static int bridge_netdev_open(struct net_device *net_dev)
{
	struct orinoco_private *priv = net_dev->priv;
	struct bridge_priv *dev = priv->card;
	int retval;

	retval = dev->orinoco_open(net_dev);
	if(watchdog_secs)
		mod_timer(&dev->watchdog_timer, jiffies+watchdog_secs*HZ);
	return retval;
}

static inline void
bridge_dump_data (const char *function, const void *data_p, int size)
{
	int i;
	const u8 *data = data_p;

	if (!debug)
		return;
	
	printk (KERN_DEBUG __FILE__": %s - length: %d", 
		function, size);

        printk("\n"KERN_DEBUG"        ");
        for (i=0; i<0x10; i++)
                printk("%02X ", i);
        for (i=0; i<size; i++){
                if((i%0x10) == 0)
                        printk("\n"KERN_DEBUG"0x%04X: ", i/0x10);
                printk("%02X ", data[i]);
        }
        printk("\n");


}


static inline void bridge_delete (struct bridge_priv *dev)
{
	//struct net_device *net_dev = dev->net_dev;
	//struct orinoco_private *priv = net_dev->priv;
	
	dbg("ENTER");
	if(in_interrupt()){
		err("orinoco_usb: %s: In interrupt", __FUNCTION__);
		return;
	}
	if(!dev){
		err("%s: dev == NULL", __FUNCTION__);
		return;
	}
	del_timer_sync(&dev->watchdog_timer);
	if (dev->net_dev)
		unregister_netdev(dev->net_dev);
	kfree (dev->read.urb->transfer_buffer);
	if (dev->bap.buf != NULL)
		kfree (dev->bap.buf);
	if (dev->read.urb != NULL)
		usb_free_urb(dev->read.urb);
	if (dev->net_dev)
		kfree(dev->net_dev);
	dbg("EXIT");
}

static void
bridge_bulk_in_callback (struct urb *urb, struct pt_regs *regs)
{
	struct bridge_priv *dev = (struct bridge_priv *)urb->context;
	struct net_device *net_dev = dev->net_dev;
	struct orinoco_private *priv = net_dev->priv;
	hermes_t *hw = &priv->hw;
	struct bridge_ans_packet *ans = urb->transfer_buffer;

	if (dev->udev == NULL){
		dbg("disconnected");
		return;
	}

	if(watchdog_secs)
		mod_timer(&dev->watchdog_timer, jiffies+watchdog_secs*HZ);

	if (urb->status == -ETIMEDOUT){
		warn("%s: urb timed out, resubmiting", __FUNCTION__);
		goto resubmit;
	}
	if (urb->status == -ECONNABORTED){
		warn("%s: connection abort, resubmiting urb", __FUNCTION__);
		goto resubmit;
	}
	if ((urb->status == -EILSEQ)
	    || (urb->status == -ENOENT)){
		dbg("status %d, not resubmiting", urb->status);
		goto exit;
	}
	if(urb->status)
		dbg("status: %d length: %d",
		    urb->status, urb->actual_length);
	if (urb->actual_length < sizeof(*ans)){
		err("%s: short read, ignoring", __FUNCTION__);
		goto resubmit;
	}

	switch (ans->hermes_rid) {
	case BRIDGE_RID_RX:
		ndbg("got a data frame");
		bridge_or_reg(hw, HERMES_EVSTAT, HERMES_EV_RX);
		orinoco_interrupt(net_dev->irq, net_dev, NULL);
		break;
	case HERMES_INQ_LINKSTATUS:
	case HERMES_INQ_TALLIES:
	case HERMES_INQ_SCAN:
		ndbg("got an INQ frame %s", rid_name(ans->hermes_rid));
#if 1
		ndump(urb->transfer_buffer, urb->actual_length);
		bridge_or_reg(hw, HERMES_EVSTAT, HERMES_EV_INFO);
		hermes_write_regn(hw, INFOFID, ans->hermes_rid);
		orinoco_interrupt(net_dev->irq, net_dev, NULL);
#endif
		break;
	default:
		if (bridge_request_in_callback(dev, urb) == -ENOENT)
			err("%s: got unexpected RID: 0x%04X(%s)", __FUNCTION__,
			    ans->hermes_rid, rid_name(ans->hermes_rid));
	}
resubmit:
	//if ((dev->net_dev->flags&IFF_UP) || !dev->read.done){
	if (dev->udev){
		ndbg("resubmiting IN urb");
		bridge_submit_in_urb(dev, NULL);
	} else {
		dbg("NOT resubmiting IN urb");
	}
exit:
	return;
}
static void watchdog_timer_func(u_long data)
{
	struct bridge_priv *dev = (struct bridge_priv *)data;
	struct net_device *net_dev = dev->net_dev;
	struct orinoco_private *priv = net_dev->priv;
	hermes_t *hw = &priv->hw;
	unsigned long flags;
	int error;

	if (!netif_running(net_dev)){
		/* Some lost timer, of course there is no traffic if the
		 * interface is not running */
		dbg("!netif_running(net_dev): ignoring timer");
		return;
	}
	if (orinoco_lock(priv, &flags) != 0) {
		dbg("called when hw_unavailable\n");
		mod_timer(&dev->watchdog_timer, jiffies+watchdog_secs*HZ);
		return;
	}
	error = hermes_inquire(hw, HERMES_INQ_TALLIES);
	orinoco_unlock(priv, &flags);
	err("card stuck? sending an INQ command to get it going (ret:%d)",
	    error);

#if 0
	/* The timer get's restarted often enough without this, and resetting
	 * it here, makes it a little too agresive, altough it does work */
	if(watchdog_secs){
		mod_timer(&dev->watchdog_timer, jiffies+watchdog_secs*HZ);
	}
#endif
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void * bridge_probe(struct usb_device *udev, unsigned int ifnum,
			 const struct usb_device_id *id)
{
	struct usb_interface *interface = &udev->actconfig->interface[ifnum];
#else
static int bridge_probe(struct usb_interface *interface,
		      const struct usb_device_id *id)
{
	struct usb_device *udev = interface_to_usbdev(interface);
#endif
	struct orinoco_private *priv;
	struct net_device *net_dev;
	hermes_t *hw;
	struct bridge_priv *dev = NULL;
	struct usb_interface_descriptor *iface_desc;
	struct usb_endpoint_descriptor *ep;
	int card_variant = -1;
	const struct firmware *fw_entry;

	int i;

	/* See if the device offered us matches what we can accept */
	
	if (usb_match_id(udev, interface, bridge_varian1_table))
		card_variant = 0;
	if (usb_match_id(udev, interface, bridge_varian2_table))
		card_variant = 1;

	if(force_unsupported)
		card_variant = -1;

	if (card_variant == -1) {
		err("Device is not supported (you may want to set"
		    " force_unsupported=1)");
		if(!force_unsupported)
			return_probe_fail(-ENODEV);
		err("Trying to handle device anyway as requested");
	}

	dbg("ENTER");
	if(!try_module_get(THIS_MODULE)){
		err("module get failed");
		return_probe_fail(-ENODEV);
	}
	down (&minor_table_mutex);

	net_dev = alloc_orinocodev(sizeof(*dev), NULL);
	if (!net_dev){
		err("Couldn't allocate net_dev");
		goto exit;
	}
	
	
	priv = net_dev->priv;
	hw = &priv->hw;
	
	dev = priv->card;
	dev->net_dev = net_dev;


	init_MUTEX (&dev->sem);
	dev->watchdog_timer.function = watchdog_timer_func;
	dev->watchdog_timer.data = (u_long)dev;
	init_timer(&dev->watchdog_timer);

	spin_lock_init(&dev->ctxq.lock);
	INIT_LIST_HEAD(&dev->ctxq.pending);
	INIT_LIST_HEAD(&dev->ctxq.active);

	dev->udev = udev;
	dev->interface = interface;

	hermes_struct_init(hw, (ulong)&dev->hermes_reg_fake,
			   HERMES_MEM, HERMES_16BIT_REGSPACING);
	hw->priv = dev;

	hw->read_reg = bridge_read_reg;
	hw->write_reg = bridge_write_reg;
	hw->init = bridge_init;
	hw->docmd_wait = bridge_docmd_wait;
	hw->xmit = bridge_xmit;
	hw->allocate = bridge_allocate;
	hw->read_ltv = bridge_read_ltv;
	hw->write_ltv = bridge_write_ltv;
	hw->bap_pread = bridge_bap_pread;
	hw->bap_pwrite = bridge_bap_pwrite;
	//net_dev->init = orinoco_init;
	dev->orinoco_stop = net_dev->stop;
	net_dev->stop = bridge_netdev_stop;
	dev->orinoco_open = net_dev->open;
	net_dev->open = bridge_netdev_open;
	//net_dev->hard_start_xmit = bridge_xmit;
	//net_dev->set_multicast_list = NULL;

#if 0
	net_dev->base_addr = (ulong)&dev->hermes_reg_fake;
	net_dev->irq = -1;
#endif
	SET_MODULE_OWNER(net_dev);
	strcpy(net_dev->name, "wlan%d");

	/* set up the endpoint information */
	/* check out the endpoints */

	iface_desc = get_iface_desc(interface, 0);
	for (i = 0; i < iface_desc->bNumEndpoints; ++i) {
		ep = get_iface_ep_desc(interface, 0, i);

		if (((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
				== USB_DIR_IN) &&
		    ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		     		== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk in endpoint */
			if (dev->read.urb != NULL) {
				warn("Found a second bulk in ep, ignored");
				continue;
			}

			dev->read.urb = usb_alloc_urb(0, GFP_KERNEL);
			if (!dev->read.urb) {
				err("No free urbs available");
				goto error;
			}
			if (ep->wMaxPacketSize != 64)
				warn("bulk in: wMaxPacketSize!= 64");
			dev->read.pipe = ep->bEndpointAddress;
			if (dev->read.pipe != (2|USB_DIR_IN))
				warn("bulk in: bEndpointAddress: %d",
						dev->read.pipe);
			dev->read.urb->transfer_buffer =
				kmalloc (BULK_BUF_SIZE, GFP_KERNEL);
			if (!dev->read.urb->transfer_buffer) {
				err("Couldn't allocate IN buffer");
				goto error;
			}
			usb_fill_bulk_urb(
				dev->read.urb, udev, 
				usb_sndbulkpipe(udev, ep->bEndpointAddress),
				dev->read.urb->transfer_buffer, BULK_BUF_SIZE,
				bridge_bulk_in_callback, dev);
		}
		
		if (((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
				== USB_DIR_OUT) &&
		    ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		     		== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk out endpoint */
			if (dev->bap.buf != NULL) {
				warn("Found a second bulk out ep, ignored");
				continue;
			}

			if (ep->wMaxPacketSize != 64)
				warn("bulk out: wMaxPacketSize != 64");
			dev->write.pipe = ep->bEndpointAddress;
			if (dev->write.pipe != 2)
				warn("bulk in: bEndpointAddress: %d",
						dev->read.pipe);
			dev->bap.buf = kmalloc (BULK_BUF_SIZE, GFP_KERNEL);
			if (!dev->bap.buf) {
				err("Couldn't allocate bulk_out_buffer");
				goto error;
			}
		}
	}
	if (!dev->bap.buf || !dev->read.urb){
		err("Didn't find the required bulk endpoints");
		goto error;
	}

	if(card_variant != -1)
		dev->card_variant = card_variant;
	else
		dev->card_variant = default_card_variant;

	/* net_dev->name is still "wlan%d", but the default hotplug
	 * support doesn't really need DEVIDE anyway, drivers that
	 * need it could use dev_alloc_name/register_netdevice instead
	 * of register_netdev */
	if (request_firmware(&fw_entry, "orinoco_ezusb_fw",
			     fw_dev_param(dev))==0) {
		firmware.size = fw_entry->size;
		firmware.code = fw_entry->data;
		find_fw_variant_offset(&firmware);
	}
	if (firmware.size && firmware.code) {
		bridge_firmware_download(dev, &firmware);
	} else {
		err("No firmware to download");
		goto error;
	}

	/* If the firmware is already downloaded orinoco.c will call
	 * bridge_init but if the firmware is not already there, that will make
	 * the kernel very unstable, so we try initializing here and quit in
	 * case of error */
	if (bridge_init(hw) < 0){
		err("Couldn't initialize the device");
		err("Firmware may not be downloaded or may be wrong.");
		if(card_variant == -1){
			err("Replug the device to pretend it is the other"
			    " variant");
			default_card_variant = !default_card_variant;
		}
		bridge_remove_in_urb(dev);
		goto error;
	}
	if (card_variant == -1){
		err("Your device variant seams to be 0x%02X, please send this"
		    " message and the output of 'lsusb -v' to"
		    " orinoco-usb-devel@ranty.ddts.net", dev->card_variant);
	}

	ndbg("registering net_dev");
	if (register_netdev(net_dev) != 0) {
		err("%s: register_netdev() failed", __FUNCTION__);
		goto error;
	}
	ndbg("net_dev registered");

	/* let the user know what node this device is now attached to */
	info ("Orinoco USB device now attached %s", net_dev->name);
	goto exit;
	
error:
	bridge_delete (dev);
	dev = NULL;
	module_put(THIS_MODULE);

exit:
	if(fw_entry){
		firmware.code = NULL;
		firmware.size = 0;
		release_firmware(fw_entry);
	}
	up (&minor_table_mutex);
	dbg("EXIT");
	usb_set_intfdata (interface, dev);
	return_probe_ok(dev);
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void bridge_disconnect(struct usb_device *udev, void *ptr)
{
	struct bridge_priv *dev = ptr;
#else
static void bridge_disconnect(struct usb_interface *intf)
{
	struct bridge_priv *dev = usb_get_intfdata (intf);
#endif
	struct net_device *net_dev = NULL;
	struct orinoco_private *priv = NULL;
	int retval;

	usb_set_intfdata (intf, NULL);

	if(!dev){
		dbg("NULL device pointer");
		return;
	}
	net_dev = dev->net_dev;
	priv = net_dev->priv;
	down (&minor_table_mutex);
	down (&dev->sem);


	if(in_interrupt()){
		dbg("in_interrupt");
	}
	//priv->hw_unavailable = 1;

	if(dev->read.urb->status == -EINPROGRESS){
		warn("%s: Unlinking pending IN urb", __FUNCTION__);
		retval = bridge_remove_in_urb(dev);
		if(retval){
			dbg("retval %d status %d", retval,
			    dev->read.urb->status);
		}
	}
	WARN_ON(!list_empty(&dev->ctxq.active)
		|| !list_empty(&dev->ctxq.pending));

	if (dev->read.urb->status == -EINPROGRESS) {
		err("orinoco_usb: Some URB in progress");
	}
	dev->udev = NULL;
	up (&dev->sem);
	bridge_delete (dev);

	info("Orinoco USB device now disconnected");
	up (&minor_table_mutex);
	module_put(THIS_MODULE);
}

static int __init bridge_module_init(void)
{
	int result;

	if(firmware.size){
		find_fw_variant_offset(&firmware);
		register_firmware("orinoco_ezusb_fw", firmware.code,
				  firmware.size);
	}
	/* register this driver with the USB subsystem */
	result = usb_register(&orinoco_driver);
	if (result < 0) {
		err("usb_register failed for the "
		    __FILE__" driver. Error number %d",
		    result);
		return -1;
	}

	printk(KERN_INFO "orinoco_usb.c " DRIVER_VERSION
	       " (Manuel Estrada Sainz <ranty@debian.org>)\n");
	return 0;
}


static void __exit bridge_module_exit(void)
{
	/* deregister this driver with the USB subsystem */
	usb_deregister(&orinoco_driver);
}


module_init (bridge_module_init);
module_exit (bridge_module_exit);

MODULE_AUTHOR("Manuel Estrada Sainz <ranty@debian.org>");
MODULE_DESCRIPTION("Orinoco USB Bridge Driver");
MODULE_LICENSE("GPL");

